"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDefSymbol = exports.getExistingNames = exports.getNamedValues = exports.isInValue = exports.isDirective = exports.extractJsModifierReturnType = exports.extractTsSignature = exports.ProviderLocation = exports.fixAndProcess = exports.createMeta = exports.getRefs = exports.getRenameRefs = exports.Provider = void 0;
const path_1 = __importStar(require("path"));
const typescript_1 = __importDefault(require("typescript"));
const postcss = __importStar(require("postcss"));
const postcss_value_parser_1 = __importDefault(require("postcss-value-parser"));
const css_selector_tokenizer_1 = __importDefault(require("css-selector-tokenizer"));
const core_1 = require("@stylable/core");
const index_internal_1 = require("@stylable/core/dist/index-internal");
const vscode_uri_1 = require("vscode-uri");
const completion_providers_1 = require("./completion-providers");
const completion_types_1 = require("./completion-types");
const pseudo_class_1 = require("./feature/pseudo-class");
const postcss_ast_utils_1 = require("./utils/postcss-ast-utils");
const selector_analyzer_1 = require("./utils/selector-analyzer");
function findLast(arr, predicate) {
    for (let index = arr.length - 1; index >= 0; index--) {
        const item = arr[index];
        if (predicate(item, index, arr)) {
            return item;
        }
    }
    return null;
}
class Provider {
    constructor(stylable, tsLangService) {
        this.stylable = stylable;
        this.tsLangService = tsLangService;
        this.providers = [
            completion_providers_1.RulesetInternalDirectivesProvider,
            completion_providers_1.ImportInternalDirectivesProvider,
            completion_providers_1.TopLevelDirectiveProvider,
            completion_providers_1.ValueDirectiveProvider,
            completion_providers_1.GlobalCompletionProvider,
            completion_providers_1.SelectorCompletionProvider,
            completion_providers_1.ExtendCompletionProvider,
            completion_providers_1.CssMixinCompletionProvider,
            completion_providers_1.CodeMixinCompletionProvider,
            completion_providers_1.FormatterCompletionProvider,
            completion_providers_1.NamedCompletionProvider,
            completion_providers_1.StateTypeCompletionProvider,
            completion_providers_1.StateSelectorCompletionProvider,
            completion_providers_1.StateEnumCompletionProvider,
            completion_providers_1.PseudoElementCompletionProvider,
            completion_providers_1.ValueCompletionProvider,
            completion_providers_1.StImportNamedCompletionProvider,
        ];
    }
    provideCompletionItemsFromSrc(src, pos, fileName, fs) {
        const res = fixAndProcess(src, pos, fileName);
        const completions = [];
        if (!res.processed.meta) {
            return [];
        }
        const options = this.createProviderOptions(src, pos, res.processed.meta, res.processed.fakes, res.currentLine, res.cursorLineIndex, fs);
        for (const provider of this.providers) {
            completions.push(...provider.provide(options));
        }
        return this.dedupeComps(completions);
    }
    getDefinitionLocation(src, position, filePath, fs) {
        if (!filePath.endsWith('.st.css')) {
            return [];
        }
        const callingMeta = this.stylable.analyze(filePath);
        const { word, meta } = getDefSymbol(src, position, filePath, this.stylable);
        if (!meta || !word) {
            return [];
        }
        const defs = [];
        let temp = null;
        let stateMeta;
        let maybeRequestPath;
        try {
            maybeRequestPath = this.stylable.resolver.resolvePath(fs.dirname(meta.source), word);
        }
        catch {
            // todo: figure out proper logging
        }
        if (maybeRequestPath) {
            if (fs.statSync(maybeRequestPath)) {
                defs.push(new ProviderLocation(maybeRequestPath, {
                    start: { line: 0, character: 0 },
                    end: { line: 0, character: 0 },
                }));
            }
        }
        else if (Object.keys(meta.getAllSymbols()).find((sym) => sym === word.replace('.', ''))) {
            const symbol = meta.getSymbol(word.replace('.', ''));
            switch (symbol._kind) {
                case 'class': {
                    defs.push(new ProviderLocation(meta.source, this.findWord(word.replace('.', ''), fs.readFileSync(meta.source, 'utf8'), position)));
                    break;
                }
                case 'var': {
                    defs.push(new ProviderLocation(meta.source, this.findWord(word, fs.readFileSync(meta.source, 'utf8'), position)));
                    break;
                }
                case 'import': {
                    let resolved = null;
                    try {
                        resolved = this.stylable.resolver.resolve(symbol);
                    }
                    catch {
                        /**/
                    }
                    let filePath;
                    if (resolved && resolved._kind !== 'js') {
                        filePath = resolved.meta.source;
                    }
                    else {
                        try {
                            filePath = this.stylable.resolvePath((0, path_1.dirname)(meta.source), symbol.import.request);
                        }
                        catch {
                            // todo: figure out proper logging
                        }
                    }
                    if (filePath) {
                        const doc = fs.readFileSync(filePath, 'utf8');
                        if (doc !== '') {
                            defs.push(new ProviderLocation(filePath, this.findWord(word, doc, position)));
                        }
                    }
                    break;
                }
            }
        }
        else if (Object.values(meta.getAllSymbols()).some((k) => {
            if (k._kind === 'class') {
                const symbolStates = k[`-st-states`];
                if (symbolStates && Object.keys(symbolStates).some((key) => key === word)) {
                    const postcsspos = new completion_providers_1.ProviderPosition(position.line + 1, position.character);
                    const pfp = (0, postcss_ast_utils_1.pathFromPosition)(callingMeta.rawAst, postcsspos, [], true);
                    const selec = pfp[pfp.length - 1].selector;
                    // If called from -st-state, i.e. inside node, pos is not in selector.
                    // Use 1 and not 0 for selector that starts with'.'
                    const char = (0, postcss_ast_utils_1.isInNode)(postcsspos, pfp[pfp.length - 1])
                        ? 1
                        : position.character;
                    const parsel = (0, selector_analyzer_1.parseSelector)(selec, char);
                    const t = parsel.target;
                    const arr = Array.isArray(t.focusChunk)
                        ? t.focusChunk[t.index].text
                        : t.focusChunk.text;
                    let name = findLast(arr, (str) => !str.startsWith(':') || str.startsWith('::'));
                    name = name.replace('.', '').replace(/:/g, '');
                    const localSymbol = callingMeta.getSymbol(name);
                    if (name === k.name ||
                        (!name.startsWith(name.charAt(0).toLowerCase()) && k.name === 'root')) {
                        temp = k;
                        stateMeta = meta;
                        return true;
                    }
                    else if (localSymbol &&
                        (localSymbol._kind === 'class' || localSymbol._kind === 'element') &&
                        localSymbol[`-st-extends`]) {
                        const res = this.findMyState(callingMeta, name, word);
                        if (res) {
                            temp = k;
                            stateMeta = res.meta;
                            return true;
                        }
                    }
                }
            }
            return false;
        })) {
            if (temp) {
                /* This is here because typescript does not recognize size effects during the if statement */
                // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-assertion
                const name = temp.name;
                defs.push(new ProviderLocation(meta.source, this.findWord(name, fs.readFileSync(stateMeta.source, 'utf8'), position)));
            }
        }
        else if (Object.keys(meta.customSelectors).find((sym) => sym === ':--' + word)) {
            defs.push(new ProviderLocation(meta.source, this.findWord(':--' + word, src, position)));
        }
        else if (!word.startsWith(word.charAt(0).toLowerCase())) {
            // Default import, link to top of imported stylesheet
            defs.push(new ProviderLocation(meta.source, (0, completion_providers_1.createRange)(0, 0, 0, 0)));
        }
        return defs;
    }
    findMyState(origMeta, elementName, state) {
        const importedSymbol = origMeta.getClass(elementName)[`-st-extends`];
        let res = null;
        if (importedSymbol && importedSymbol._kind === 'import') {
            res = this.stylable.resolver.resolveImport(importedSymbol);
        }
        const localSymbol = origMeta.getSymbol(elementName);
        if (res &&
            res._kind === 'css' &&
            Object.keys(res.symbol[`-st-states`]).includes(state)) {
            return res;
        }
        else if (res &&
            res._kind === 'css' &&
            (localSymbol._kind === 'class' || localSymbol._kind === 'element') &&
            localSymbol[`-st-extends`]) {
            return this.findMyState(res.meta, res.symbol.name, state);
        }
        else {
            return null;
        }
    }
    getSignatureHelp(src, pos, filePath, fs, paramInfo) {
        if (!filePath.endsWith('.st.css')) {
            return null;
        }
        const { processed: { meta }, } = fixAndProcess(src, pos, filePath);
        if (!meta) {
            return null;
        }
        const split = src.split('\n');
        const line = split[pos.line];
        let value = '';
        const stPath = (0, postcss_ast_utils_1.pathFromPosition)(meta.rawAst, {
            line: pos.line + 1,
            character: pos.character + 1,
        });
        const lastStPath = stPath[stPath.length - 1];
        if ((0, postcss_ast_utils_1.isRoot)(lastStPath)) {
            return this.getSignatureForStateWithParamSelector(meta, pos, line);
        }
        else if (line.slice(0, pos.character).trim().startsWith(`-st-states`)) {
            return this.getSignatureForStateWithParamDefinition(pos, line);
        }
        // If last node is not root, we're in a declaration [TODO: or a media query]
        if (line.slice(0, pos.character).trim().startsWith(`-st-mixin`)) {
            // TODO: handle multiple lines as well
            value = line
                .slice(0, pos.character)
                .trim()
                .slice(`-st-mixin`.length + 1)
                .trim();
        }
        else if (line.slice(0, pos.character).trim().includes(':')) {
            value = line
                .slice(0, pos.character)
                .trim()
                .slice(line.slice(0, pos.character).trim().indexOf(':') + 1)
                .trim();
        }
        if (/value\(\s*[^)]*$/.test(value)) {
            return null;
        }
        const parsed = (0, postcss_value_parser_1.default)(value);
        let mixin = '';
        const rev = parsed.nodes[parsed.nodes.length - 1];
        if (rev.type === 'function' && !!rev.unclosed) {
            mixin = rev.value;
        }
        else {
            return null;
        }
        const activeParam = rev.nodes.reduce((acc, cur) => {
            return cur.type === 'div' ? acc + 1 : acc;
        }, 0);
        if (mixin === 'value') {
            return null;
        }
        const mappedSymbol = meta.getSymbol(mixin);
        if (mappedSymbol && mappedSymbol._kind === 'import') {
            if (mappedSymbol.import.from.endsWith('.ts')) {
                return this.getSignatureForTsModifier(mixin, activeParam, mappedSymbol.import.from, meta.getSymbol(mixin).type === 'default', paramInfo);
            }
            else if (mappedSymbol.import.from.endsWith('.js')) {
                if (fs.fileExistsSync(mappedSymbol.import.from.slice(0, -3) + '.d.ts')) {
                    return this.getSignatureForTsModifier(mixin, activeParam, mappedSymbol.import.from.slice(0, -3) + '.d.ts', mappedSymbol.type === 'default', paramInfo);
                }
                else {
                    const resolvedPath = this.stylable.resolvePath((0, path_1.dirname)(meta.source), mappedSymbol.import.request);
                    const fileExists = fs.fileExistsSync(resolvedPath);
                    return fileExists
                        ? this.getSignatureForJsModifier(mixin, activeParam, fs.readFileSync(resolvedPath, 'utf8'), paramInfo)
                        : null;
                }
            }
        }
        return null;
    }
    findWord(word, src, position) {
        const split = src.split('\n');
        const regex = '\\b' + '\\.?' + this.escapeRegExp(word.replace('.', '').replace(':--', '')) + '\\b';
        let lineIndex = split.findIndex((l) => {
            const reg = RegExp(regex).exec(l);
            return !!reg && l.slice(reg.index - 2, reg.index) !== '::';
        });
        if (lineIndex === -1 || lineIndex === position.line) {
            lineIndex = position.line;
            // lineIndex = split.findIndex(l => l.trim().indexOf(word) !== -1)
        }
        if (lineIndex === -1) {
            return (0, completion_providers_1.createRange)(0, 0, 0, 0);
        }
        const line = split[lineIndex];
        const match = line.match(RegExp(regex));
        if (match) {
            return (0, completion_providers_1.createRange)(lineIndex, line.lastIndexOf(word), lineIndex, line.lastIndexOf(word) + word.length);
        }
        else {
            return (0, completion_providers_1.createRange)(0, 0, 0, 0);
        }
    }
    escapeRegExp(re) {
        return re.replace(/([.*+?^=!:${}()|[\]/\\])/g, '\\$1');
    }
    getSignatureForTsModifier(mixin, activeParam, filePath, isDefault, paramInfo) {
        const sig = extractTsSignature(filePath, mixin, isDefault, this.tsLangService);
        if (!sig || !sig.declaration) {
            return null;
        }
        const ptypes = sig.parameters.map((p) => {
            return (p.name +
                ':' +
                p.valueDeclaration.type.getFullText());
        });
        const rtype = sig.declaration.type
            ? sig.declaration.type.getText()
            : '';
        const parameters = sig.parameters.map((pt) => {
            const label = pt.name +
                ':' +
                pt.valueDeclaration.type.getFullText();
            return paramInfo.create(label);
        });
        const sigInfo = {
            label: mixin + '(' + ptypes.join(', ') + '): ' + rtype,
            parameters,
        };
        return {
            activeParameter: activeParam,
            activeSignature: 0,
            signatures: [sigInfo],
        };
    }
    getSignatureForJsModifier(mixin, activeParam, fileSrc, paramInfo) {
        const lines = fileSrc.split('\n');
        const mixinLine = lines.findIndex((l) => l.trim().startsWith('exports.' + mixin));
        const docStartLine = lines.slice(0, mixinLine).lastIndexOf(lines
            .slice(0, mixinLine)
            .reverse()
            .find((l) => l.trim().startsWith('/**')));
        const docLines = lines.slice(docStartLine, mixinLine);
        const formattedLines = [];
        docLines.forEach((l) => {
            if (l.trim().startsWith('*/')) {
                return;
            }
            if (l.trim().startsWith('/**') && !!l.trim().slice(3).trim()) {
                formattedLines.push(l.trim().slice(3).trim());
            }
            if (l.trim().startsWith('*')) {
                formattedLines.push(l.trim().slice(1).trim());
            }
        });
        const returnStart = formattedLines.findIndex((l) => l.startsWith('@returns'));
        const returnEnd = formattedLines.slice(returnStart + 1).findIndex((l) => l.startsWith('@')) === -1
            ? formattedLines.length - 1
            : formattedLines.slice(returnStart + 1).findIndex((l) => l.startsWith('@')) +
                returnStart;
        const returnLines = formattedLines.slice(returnStart, returnEnd + 1);
        formattedLines.splice(returnStart, returnLines.length);
        const returnType = /@returns *{(\w+)}/.exec(returnLines[0])
            ? /@returns *{(\w+)}/.exec(returnLines[0])[1]
            : '';
        const summaryStart = formattedLines.findIndex((l) => l.startsWith('@summary'));
        let summaryLines = [];
        if (summaryStart !== -1) {
            const summaryEnd = formattedLines.slice(summaryStart + 1).findIndex((l) => l.startsWith('@')) === -1
                ? formattedLines.length - 1
                : formattedLines.slice(summaryStart + 1).findIndex((l) => l.startsWith('@')) +
                    summaryStart;
            summaryLines = formattedLines.slice(summaryStart, summaryEnd + 1);
            formattedLines.splice(summaryStart, summaryLines.length);
        }
        const params = [];
        while (formattedLines.find((l) => l.startsWith('@param'))) {
            const paramStart = formattedLines.findIndex((l) => l.startsWith('@param'));
            const paramEnd = formattedLines.slice(paramStart + 1).findIndex((l) => l.startsWith('@')) === -1
                ? formattedLines.length - 1
                : formattedLines.slice(paramStart + 1).findIndex((l) => l.startsWith('@')) +
                    paramStart;
            const paramLines = formattedLines.slice(paramStart, paramEnd + 1);
            formattedLines.splice(paramStart, paramLines.length);
            if (/@param *{([ \w<>,'"|]*)} *(\w*)(.*)/.exec(paramLines[0])) {
                params.push([
                    /@param *{([ \w<>,'"|]*)} *(\w*)(.*)/.exec(paramLines[0])[1],
                    /@param *{([ \w<>,'"|]*)} *(\w*)(.*)/.exec(paramLines[0])[2],
                    /@param *{([ \w<>,'"|]*)} *(\w*)(.*)/.exec(paramLines[0])[3],
                ]);
            }
        }
        let descLines = [];
        if (formattedLines.find((l) => l.startsWith('@description'))) {
            const descStart = formattedLines.findIndex((l) => l.startsWith('@description'));
            const descEnd = formattedLines.slice(descStart + 1).findIndex((l) => l.startsWith('@')) === -1
                ? formattedLines.length - 1
                : formattedLines.slice(descStart + 1).findIndex((l) => l.startsWith('@')) +
                    descStart;
            descLines = formattedLines.slice(descStart, descEnd + 1);
        }
        else if (formattedLines.findIndex((l) => l.startsWith('@')) === -1) {
            descLines = formattedLines;
        }
        else {
            descLines = formattedLines.slice(0, formattedLines.findIndex((l) => l.startsWith('@')) + 1);
        }
        if (descLines[0] && descLines[0].startsWith('@description')) {
            descLines[0] = descLines[0].slice(12).trim();
        }
        const parameters = params.map((p) => paramInfo.create(p[1] + ': ' + p[0], p[2].trim()));
        const sigInfo = {
            label: mixin + '(' + parameters.map((p) => p.label).join(', ') + '): ' + returnType,
            documentation: descLines.join('\n'),
            parameters,
        };
        return {
            activeParameter: activeParam,
            activeSignature: 0,
            signatures: [sigInfo],
        };
    }
    getSignatureForStateWithParamSelector(meta, pos, line) {
        let word = '';
        const posChar = pos.character + 1;
        const parsed = css_selector_tokenizer_1.default.parse(line);
        if (parsed.nodes[0].type === 'selector') {
            let length = 0;
            parsed.nodes[0].nodes.forEach((node) => {
                if (node.type === 'invalid') {
                    return; // TODO: refactor - handles places outside of a selector
                }
                else if (node.type === 'spacing') {
                    length += node.value.length;
                }
                else if (node.name !== undefined) {
                    length += node.name.length + 1;
                    if (node.type === 'pseudo-class' &&
                        posChar > length + 1 &&
                        node.content !== undefined &&
                        posChar <= length + 2 + node.content.length) {
                        word = node.name;
                    }
                }
            });
        }
        let stateDef = null;
        if (word) {
            const resolvedElements = this.stylable.transformSelector(meta, line).resolved;
            resolvedElements[0][0].resolved.forEach((el) => {
                const symbolStates = el.symbol[`-st-states`];
                if (symbolStates && typeof symbolStates[word] === 'object') {
                    stateDef = symbolStates[word];
                }
            });
            if (stateDef && typeof stateDef === 'object') {
                const parameters = (0, pseudo_class_1.resolveStateParams)(stateDef);
                const sigInfo = {
                    label: `${word}(${parameters})`,
                    parameters: [{ label: parameters }],
                };
                return {
                    activeParameter: 0,
                    activeSignature: 0,
                    signatures: [sigInfo],
                };
            }
        }
        return null;
    }
    getSignatureForStateWithParamDefinition(pos, line) {
        const res = (0, pseudo_class_1.resolveStateTypeOrValidator)(pos, line);
        if (typeof res === 'string') {
            return (0, pseudo_class_1.createStateValidatorSignature)(res);
        }
        else if (typeof res === 'boolean') {
            return (0, pseudo_class_1.createStateTypeSignature)();
        }
        else {
            return null;
        }
    }
    createProviderOptions(src, position, meta, fakeRules, fullLineText, cursorPosInLine, fs) {
        const path = (0, postcss_ast_utils_1.pathFromPosition)(meta.rawAst, {
            line: position.line + 1,
            character: position.character,
        });
        const astAtCursor = path[path.length - 1];
        const parentAst = astAtCursor.parent
            ? astAtCursor.parent
            : undefined;
        const parentSelector = parentAst &&
            (0, postcss_ast_utils_1.isSelector)(parentAst) &&
            fakeRules.findIndex((f) => {
                return f.selector === parentAst.selector;
            }) === -1
            ? parentAst
            : astAtCursor &&
                (0, postcss_ast_utils_1.isSelector)(astAtCursor) &&
                fakeRules.findIndex((f) => {
                    return f.selector === astAtCursor.selector;
                }) === -1
                ? astAtCursor
                : null;
        const { lineChunkAtCursor, fixedCharIndex } = getChunkAtCursor(fullLineText, cursorPosInLine);
        const ps = (0, selector_analyzer_1.parseSelector)(lineChunkAtCursor, fixedCharIndex);
        const chunkStrings = ps.selector.reduce((acc, s) => {
            return acc.concat(s.text);
        }, []);
        const currentSelector = ps.selector[0].classes[0] ||
            ps.selector[0].customSelectors[0] ||
            chunkStrings[0];
        const expandedLine = (0, core_1.expandCustomSelectors)(postcss.rule({ selector: lineChunkAtCursor }), meta.customSelectors)
            .split(' ')
            .pop(); // TODO: replace with selector parser
        const resolvedElements = this.stylable.transformSelector(meta, expandedLine).resolved;
        const resolvedRoot = this.stylable.transformSelector(meta, `.${meta.root}`).resolved[0][0];
        let resolved = [];
        if (currentSelector && resolvedElements[0].length) {
            const clas = resolvedElements[0].find((e) => e.type === 'class' || (e.type === 'element' && e.resolved.length > 1)); // TODO: better type parsing
            resolved = clas ? clas.resolved : [];
        }
        return {
            meta,
            fs,
            stylable: this.stylable,
            src,
            tsLangService: this.tsLangService,
            resolvedElements,
            resolvedRoot,
            parentSelector,
            astAtCursor,
            lineChunkAtCursor,
            lastSelectoid: ps.lastSelector,
            fullLineText,
            position,
            resolved,
            currentSelector,
            target: ps.target,
            isMediaQuery: isInMediaQuery(path),
            fakes: fakeRules,
        };
    }
    dedupeComps(completions) {
        const uniqs = new Map();
        completions.forEach((comp) => {
            if (!uniqs.has(comp.label)) {
                uniqs.set(comp.label, comp);
            }
        });
        const res = [];
        uniqs.forEach((v) => res.push(v));
        return res;
    }
}
exports.Provider = Provider;
function isIllegalLine(line) {
    return /^\s*[-.:]+\s*$/.test(line);
}
const lineEndsRegexp = /({|}|;)/;
function findRefs(word, defMeta, scannedMeta, callingMeta, stylable, pos) {
    if (!word) {
        return [];
    }
    const refs = [];
    if (word.startsWith(':global(')) {
        scannedMeta.rawAst.walkRules((rule) => {
            if (rule.selector.includes(word) && rule.source && rule.source.start) {
                refs.push({
                    uri: vscode_uri_1.URI.file(scannedMeta.source).toString(),
                    range: {
                        start: {
                            line: rule.source.start.line - 1,
                            character: rule.selector.indexOf(word) + ':global('.length,
                        },
                        end: {
                            line: rule.source.start.line - 1,
                            character: rule.selector.indexOf(word) + word.length - 1,
                        },
                    },
                });
            }
        });
        return refs;
    }
    const valueRegex = new RegExp('(\\.?' + word + ')(\\s|$|\\:|;|\\)|,)', 'g');
    scannedMeta.rawAst.walkRules((rule) => {
        // Usage in selector
        const filterRegex = new RegExp('(\\.?' + word + ')(\\s|$|\\:|;|\\))', 'g');
        if (filterRegex.test(rule.selector) && !!rule.source && !!rule.source.start) {
            const resScanned = stylable.transformSelector(scannedMeta, rule.selector).resolved;
            if (resScanned[0].some((rl) => {
                return (rl.name === word &&
                    rl.resolved.some((i) => i.meta.source === defMeta.source));
                // return rl.name === word && last(rl.resolved)!.meta.source === defMeta.source
            })) {
                let match = valueRegex.exec(rule.selector);
                while (match !== null) {
                    const index = match[0].startsWith('.') ? match.index : match.index - 1;
                    refs.push({
                        uri: vscode_uri_1.URI.file(scannedMeta.source).toString(),
                        range: {
                            start: {
                                line: rule.source.start.line - 1,
                                character: rule.source.start.column + index,
                            },
                            end: {
                                line: rule.source.start.line - 1,
                                character: rule.source.start.column + index + word.length,
                            },
                        },
                    });
                    match = valueRegex.exec(rule.selector);
                }
            }
            else if (!!pos &&
                resScanned[0].some((rs) => {
                    const postcsspos = new completion_providers_1.ProviderPosition(pos.line + 1, pos.character);
                    const pfp = (0, postcss_ast_utils_1.pathFromPosition)(callingMeta.rawAst, postcsspos, [], true);
                    let lastStPath = pfp[pfp.length - 1];
                    if (lastStPath.type === 'decl') {
                        lastStPath = pfp[pfp.length - 2];
                    }
                    const char = (0, postcss_ast_utils_1.isInNode)(postcsspos, pfp[pfp.length - 1]) ? 1 : pos.character;
                    const callPs = (0, selector_analyzer_1.parseSelector)(lastStPath.selector, char);
                    const callingElement = findLast(callPs.selector[callPs.target.index].text.slice(0, callPs.target.internalIndex + 1), (e) => !e.startsWith(':') || e.startsWith('::'));
                    if (!callingElement) {
                        return false;
                    }
                    const selector = lastStPath.selector;
                    const selectorElement = stylable.transformSelector(callingMeta, selector.slice(0, selector.indexOf(word) + word.length)).resolved[0];
                    const resolvedSelectorElement = selectorElement[selectorElement.length - 1].resolved;
                    const lastResolvedSelector = resolvedSelectorElement[resolvedSelectorElement.length - 1];
                    if (rs.resolved.some((inner) => inner.meta.source === defMeta.source &&
                        Object.keys(inner.symbol[`-st-states`]).includes(word)) &&
                        rs.resolved[rs.resolved.length - 1].symbol.name ===
                            lastResolvedSelector.symbol.name) {
                        return true;
                    }
                    return false;
                })) {
                let match = valueRegex.exec(rule.selector);
                while (match !== null) {
                    const index = match[0].startsWith('.') ? match.index : match.index - 1;
                    refs.push({
                        uri: vscode_uri_1.URI.file(scannedMeta.source).toString(),
                        range: {
                            start: {
                                line: rule.source.start.line - 1,
                                character: rule.source.start.column + index,
                            },
                            end: {
                                line: rule.source.start.line - 1,
                                character: rule.source.start.column + index + word.length,
                            },
                        },
                    });
                    match = valueRegex.exec(rule.selector);
                }
            }
        }
    });
    scannedMeta.rawAst.walkDecls((decl) => {
        if (!decl.source || !decl.source.start) {
            return;
        }
        const directiveRegex = new RegExp(`-st-extends` + '|' + `-st-named` + '|' + `-st-default`);
        if (directiveRegex.test(decl.prop)) {
            // Usage in -st directives
            const reg = new RegExp(valueRegex.source);
            const match = reg.exec(decl.value);
            if (match) {
                refs.push({
                    uri: vscode_uri_1.URI.file(scannedMeta.source).toString(),
                    range: {
                        start: {
                            line: decl.source.start.line - 1,
                            character: match.index +
                                decl.source.start.column +
                                decl.prop.length +
                                (decl.raws.between ? decl.raws.between.length : 0) -
                                1,
                        },
                        end: {
                            line: decl.source.start.line - 1,
                            character: match.index +
                                decl.source.start.column +
                                decl.prop.length +
                                (decl.raws.between ? decl.raws.between.length : 0) +
                                word.length -
                                1,
                        },
                    },
                });
            }
        }
    });
    scannedMeta.rawAst.walkDecls((decl) => {
        if (!decl.source || !decl.source.start || !pos) {
            return;
        }
        const directiveRegex = new RegExp(`-st-states`);
        const postcsspos = new completion_providers_1.ProviderPosition(pos.line + 1, pos.character);
        const pfp = (0, postcss_ast_utils_1.pathFromPosition)(callingMeta.rawAst, postcsspos, [], true);
        const char = (0, postcss_ast_utils_1.isInNode)(postcsspos, pfp[pfp.length - 1]) ? 1 : pos.character;
        const callPs = (0, selector_analyzer_1.parseSelector)(pfp[pfp.length - 1].selector, char);
        const callingElement = findLast(callPs.selector[callPs.target.index].text.slice(0, callPs.target.internalIndex + 1), (e) => !e.startsWith(':') || e.startsWith('::'));
        const blargh = stylable.transformSelector(callingMeta, pfp[pfp.length - 1].selector).resolved;
        if (directiveRegex.test(decl.prop) &&
            scannedMeta.source === defMeta.source &&
            !!blargh.length &&
            !!callingElement &&
            blargh[0].some((inner) => {
                return (inner.name === callingElement.replace(/:/g, '').replace('.', '') &&
                    inner.resolved.some((s) => s.symbol.name ===
                        decl.parent.selector.replace('.', '')));
            })) {
            const reg = new RegExp(valueRegex.source);
            const match = reg.exec(decl.value);
            if (match) {
                refs.push({
                    uri: vscode_uri_1.URI.file(scannedMeta.source).toString(),
                    range: {
                        start: {
                            line: decl.source.start.line - 1,
                            character: match.index +
                                decl.source.start.column +
                                decl.prop.length +
                                (decl.raws.between ? decl.raws.between.length : 0) -
                                1,
                        },
                        end: {
                            line: decl.source.start.line - 1,
                            character: match.index +
                                decl.source.start.column +
                                decl.prop.length +
                                (decl.raws.between ? decl.raws.between.length : 0) +
                                word.length -
                                1,
                        },
                    },
                });
            }
        }
    });
    scannedMeta.rawAst.walkDecls(`-st-mixin`, (decl) => {
        // usage in -st-mixin
        if (!decl.source || !decl.source.start) {
            return;
        }
        const lines = decl.value.split('\n');
        lines.forEach((line, index) => {
            let match = valueRegex.exec(line);
            while (match !== null) {
                refs.push({
                    uri: vscode_uri_1.URI.file(scannedMeta.source).toString(),
                    range: {
                        start: {
                            line: decl.source.start.line - 1 + index,
                            character: index
                                ? match.index
                                : decl.source.start.column +
                                    `-st-mixin`.length +
                                    match.index +
                                    (decl.raws.between ? decl.raws.between.length : 0) -
                                    1,
                        },
                        end: {
                            line: decl.source.start.line - 1 + index,
                            character: word.length +
                                (index
                                    ? match.index
                                    : decl.source.start.column +
                                        `-st-mixin`.length +
                                        match.index +
                                        (decl.raws.between ? decl.raws.between.length : 0) -
                                        1),
                        },
                    },
                });
                match = valueRegex.exec(line);
            }
        });
    });
    scannedMeta.rawAst.walkDecls(word, (decl) => {
        // Variable definition
        if (decl.parent &&
            decl.parent.type === 'rule' &&
            decl.parent.selector === ':vars' &&
            !!decl.source &&
            !!decl.source.start) {
            refs.push({
                uri: vscode_uri_1.URI.file(scannedMeta.source).toString(),
                range: {
                    start: {
                        line: decl.source.start.line - 1,
                        character: decl.source.start.column - 1,
                    },
                    end: {
                        line: decl.source.start.line - 1,
                        character: decl.source.start.column + word.length - 1,
                    },
                },
            });
        }
    });
    scannedMeta.rawAst.walkDecls((decl) => {
        // Variable usage
        if (decl.value.includes('value(') && !!decl.source && !!decl.source.start) {
            const usageRegex = new RegExp('value\\(\\s*' + word + '\\s*\\)', 'g');
            const match = usageRegex.exec(decl.value);
            if (match) {
                refs.push({
                    uri: vscode_uri_1.URI.file(scannedMeta.source).toString(),
                    range: {
                        start: {
                            line: decl.source.start.line - 1,
                            character: match.index +
                                decl.source.start.column +
                                decl.prop.length +
                                (decl.raws.between ? decl.raws.between.length : 0) +
                                'value('.length -
                                1,
                        },
                        end: {
                            line: decl.source.start.line - 1,
                            character: match.index +
                                decl.source.start.column +
                                decl.prop.length +
                                (decl.raws.between ? decl.raws.between.length : 0) +
                                'value('.length +
                                word.length -
                                1,
                        },
                    },
                });
            }
        }
    });
    return refs;
}
function newFindRefs(word, defMeta, callingMeta, stylesheetsPath, // TODO: use docs instead?
stylable, pos) {
    let refs = [];
    if (word.startsWith(':global(')) {
        // Global selector strings are special
        stylesheetsPath.forEach((stylesheetPath) => {
            const scannedMeta = stylable.analyze(stylesheetPath);
            scannedMeta.rawAst.walkRules((rule) => {
                if (rule.selector.includes(word)) {
                    refs = refs.concat(findRefs(word, defMeta, scannedMeta, callingMeta, stylable));
                }
            });
        });
        return refs;
    }
    else {
        word = word.replace('.', '');
    }
    const defSymbols = defMeta.getAllSymbols();
    if (!defSymbols[word] && !word.startsWith(word.charAt(0).toLowerCase())) {
        // Default import
        stylesheetsPath.forEach((stylesheetPath) => {
            const scannedMeta = stylable.analyze(stylesheetPath);
            let tmp = '';
            const scannedSymbols = scannedMeta.getAllSymbols();
            if (Object.keys(scannedSymbols).some((k) => {
                tmp = k;
                const localSymbol = scannedSymbols[k];
                return ((localSymbol._kind === 'element' &&
                    localSymbol.alias &&
                    localSymbol.alias.import.from === defMeta.source) ||
                    (localSymbol._kind === 'import' &&
                        localSymbol.import.from === defMeta.source));
            })) {
                refs = refs.concat(findRefs(tmp, defMeta, scannedMeta, callingMeta, stylable));
            }
        });
    }
    else if (defSymbols[word] && defSymbols[word]._kind === 'var') {
        // Variable
        stylesheetsPath.forEach((stylesheetPath) => {
            const scannedMeta = stylable.analyze(stylesheetPath);
            const scannedSymbols = scannedMeta.getAllSymbols();
            if (!scannedSymbols[word] ||
                (scannedSymbols[word]._kind !== 'var' && scannedSymbols[word]._kind !== 'import')) {
                return;
            }
            if (scannedMeta.source === defMeta.source) {
                // We're in the defining file
                refs = refs.concat(findRefs(word.replace('.', ''), defMeta, scannedMeta, callingMeta, stylable));
            }
            else {
                // We're in a using file
                const newSymb = stylable.resolver.deepResolve(scannedSymbols[word]);
                if (!newSymb || !newSymb.meta) {
                    return;
                }
                if (newSymb.meta.source === defMeta.source) {
                    refs = refs.concat(findRefs(word.replace('.', ''), defMeta, scannedMeta, callingMeta, stylable));
                }
            }
        });
    }
    else if (defSymbols[word] &&
        (defSymbols[word]._kind === 'class' || defSymbols[word]._kind === 'import')) {
        // Elements
        const valueRegex = new RegExp('(\\.?' + word + ')\\b', 'g');
        stylesheetsPath.forEach((stylesheetPath) => {
            const scannedMeta = stylable.analyze(stylesheetPath);
            let done = false;
            scannedMeta.rawAst.walkRules((r) => {
                if (valueRegex.test(r.selector) && !done) {
                    const resolved = stylable.transformSelector(scannedMeta, r.selector).resolved;
                    const resolvedInner = resolved[0].find((r) => r.name === word);
                    if (resolvedInner &&
                        resolvedInner.resolved.some((r) => r.meta.source === defMeta.source)) {
                        refs = refs.concat(findRefs(word.replace('.', ''), defMeta, scannedMeta, callingMeta, stylable));
                        done = true;
                    }
                }
            });
            scannedMeta.rawAst.walkDecls((d) => {
                if (valueRegex.test(d.value) && !done) {
                    if (d.prop === `-st-named` &&
                        d.parent &&
                        d.parent.nodes.find((n) => {
                            return (n.prop === `-st-from` &&
                                path_1.default.resolve(path_1.default.dirname(scannedMeta.source), n.value.replace(/"/g, '')) === defMeta.source);
                        })) {
                        refs = refs.concat(findRefs(word.replace('.', ''), defMeta, scannedMeta, callingMeta, stylable));
                        done = true;
                    }
                }
            });
        });
    }
    else if (Object.values(defSymbols).some((sym) => {
        const symbolStates = sym._kind === 'class' && sym[`-st-states`];
        // states
        return (sym._kind === 'class' &&
            symbolStates &&
            Object.keys(symbolStates).some((k) => {
                if (k === word && !!pos) {
                    const postcsspos = new completion_providers_1.ProviderPosition(pos.line + 1, pos.character);
                    const pfp = (0, postcss_ast_utils_1.pathFromPosition)(callingMeta.rawAst, postcsspos, [], true);
                    let lastStPath = pfp[pfp.length - 1];
                    if (lastStPath.type === 'decl') {
                        lastStPath = pfp[pfp.length - 2];
                    }
                    // If called from -st-state, i.e. inside node, pos is not in selector.
                    const selec = lastStPath.selector;
                    // Use 1 and not 0 for selector that starts with'.'
                    const char = (0, postcss_ast_utils_1.isInNode)(postcsspos, pfp[pfp.length - 1]) ? 1 : pos.character;
                    const parsel = (0, selector_analyzer_1.parseSelector)(selec, char);
                    const t = parsel.target;
                    const arr = Array.isArray(t.focusChunk)
                        ? t.focusChunk[t.index].text
                        : t.focusChunk.text;
                    let name = findLast(arr, (str) => !str.startsWith(':') || str.startsWith('::'));
                    const pse = stylable.transformSelector(callingMeta, selec).resolved;
                    name = name.replace('.', '').replace(/:/g, '');
                    if (!!pse &&
                        !!pse[0].some((psInner) => psInner.name === name &&
                            psInner.resolved.some((r) => r.symbol.name === sym.name))) {
                        return true;
                    }
                }
                return false;
            }));
    })) {
        stylesheetsPath.forEach((stylesheetPath) => {
            const scannedMeta = stylable.analyze(stylesheetPath);
            let done = false;
            if (defMeta.source === scannedMeta.source) {
                refs = refs.concat(findRefs(word.replace('.', ''), defMeta, scannedMeta, callingMeta, stylable, pos));
                return;
            }
            if (!pos) {
                return;
            }
            scannedMeta.rawAst.walkRules((r) => {
                if (r.selector.includes(':' + word) && !done) {
                    // Won't work if word appears elsewhere in string
                    const parsed = (0, selector_analyzer_1.parseSelector)(r.selector, r.selector.indexOf(word));
                    const elem = parsed.selector[parsed.target.index].type === '*' ||
                        !parsed.selector[parsed.target.index].type.startsWith(parsed.selector[parsed.target.index].type
                            .charAt(0)
                            .toLowerCase())
                        ? findLast(parsed.selector[parsed.target.index].text, (str) => !str.startsWith(':') || str.startsWith('::')).replace('.', '')
                        : parsed.selector[parsed.target.index].name;
                    const reso = stylable.transformSelector(scannedMeta, r.selector).resolved;
                    const symb = reso[0].find((o) => o.name === elem);
                    if (!!symb &&
                        symb.resolved.some((inner) => {
                            if (inner.symbol._kind === 'class') {
                                const symbolStates = inner.symbol[`-st-states`];
                                return (symbolStates &&
                                    inner.meta.source === defMeta.source &&
                                    Object.keys(symbolStates).includes(word));
                            }
                            return false;
                        })) {
                        refs = refs.concat(findRefs(word.replace('.', ''), defMeta, scannedMeta, callingMeta, stylable, pos));
                        done = true;
                    }
                }
            });
        });
    }
    return refs;
}
function getRenameRefs(filePath, pos, fs, stylable) {
    const refs = getRefs(filePath, pos, fs, stylable);
    const newRefs = [];
    refs.forEach((ref) => {
        const FISH = vscode_uri_1.URI.parse(ref.uri).path.startsWith(stylable.projectRoot);
        const isRefInProject = ref.uri.startsWith(stylable.projectRoot);
        if (!ref.uri.includes('node_modules') && (FISH || isRefInProject)) {
            newRefs.push(ref);
        }
    });
    return newRefs;
}
exports.getRenameRefs = getRenameRefs;
function getRefs(filePath, position, fs, stylable) {
    const callingMeta = stylable.analyze(filePath);
    const symb = getDefSymbol(fs.readFileSync(filePath, 'utf8'), position, filePath, stylable);
    if (!symb.meta) {
        return [];
    }
    const stylesheets = fs
        .findFilesSync(stylable.projectRoot, {
        filterFile: (fileDesc) => {
            return fileDesc.name.endsWith('.st.css');
        },
    })
        .map((stylesheetPath) => vscode_uri_1.URI.file(stylesheetPath).fsPath);
    return newFindRefs(symb.word, symb.meta, callingMeta, stylesheets, stylable, position);
}
exports.getRefs = getRefs;
function createMeta(src, path) {
    let meta;
    const fakes = [];
    try {
        const ast = (0, index_internal_1.safeParse)(src, { from: vscode_uri_1.URI.file(path).fsPath });
        if (ast.nodes) {
            for (const node of ast.nodes) {
                if (node.type === 'decl') {
                    const r = postcss.rule({ selector: node.prop + ':' + node.value });
                    r.source = node.source;
                    node.replaceWith(r);
                    fakes.push(r);
                }
            }
        }
        if (ast.raws.after && ast.raws.after.trim()) {
            const r = postcss.rule({ selector: ast.raws.after.trim() });
            ast.append(r);
            fakes.push(r);
        }
        meta = (0, index_internal_1.process)(ast);
    }
    catch (error) {
        return { meta: null, fakes };
    }
    return {
        meta,
        fakes,
    };
}
exports.createMeta = createMeta;
function fixAndProcess(src, position, filePath) {
    let cursorLineIndex = position.character;
    const lines = src.replace(/\r\n/g, '\n').split('\n');
    let currentLine = lines[position.line];
    let fixedSrc = src;
    if (currentLine.match(lineEndsRegexp)) {
        let currentLocation = 0;
        const splitLine = currentLine.split(lineEndsRegexp);
        for (let i = 0; i < splitLine.length; i += 2) {
            currentLocation += splitLine[i].length + 1;
            if (currentLocation >= position.character) {
                currentLine = splitLine[i];
                if (isIllegalLine(currentLine)) {
                    splitLine[i] = '\n';
                    lines.splice(position.line, 1, splitLine.join(''));
                    fixedSrc = lines.join('\n');
                }
                break;
            }
            else {
                cursorLineIndex -= splitLine[i].length + 1;
            }
        }
    }
    else if (isIllegalLine(currentLine)) {
        lines.splice(position.line, 1, '');
        fixedSrc = lines.join('\n');
    }
    const processed = createMeta(fixedSrc, filePath);
    return {
        processed,
        currentLine,
        cursorLineIndex,
    };
}
exports.fixAndProcess = fixAndProcess;
class ProviderLocation {
    constructor(uri, range) {
        this.uri = uri;
        this.range = range;
    }
}
exports.ProviderLocation = ProviderLocation;
function extractTsSignature(filePath, mixin, isDefault, tsLangService) {
    tsLangService.setOpenedFiles([filePath]);
    const program = tsLangService.ts.getProgram();
    if (!program) {
        return;
    }
    const tc = program.getTypeChecker();
    const sf = program.getSourceFile(filePath);
    if (!sf) {
        return;
    }
    const mix = tc.getSymbolsInScope(sf, typescript_1.default.SymbolFlags.Function).find((f) => {
        if (isDefault) {
            return f.exportSymbol && f.exportSymbol.escapedName === 'default';
        }
        else {
            return f.exportSymbol && f.exportSymbol.escapedName === mixin;
        }
    });
    if (!mix || !mix.declarations) {
        return;
    }
    return tc.getSignatureFromDeclaration(mix.declarations[0]);
}
exports.extractTsSignature = extractTsSignature;
function extractJsModifierReturnType(mixin, fileSrc) {
    const lines = fileSrc.split('\n');
    const mixinLine = lines.findIndex((l) => l.trim().startsWith('exports.' + mixin));
    const docStartLine = lines.slice(0, mixinLine).lastIndexOf(lines
        .slice(0, mixinLine)
        .reverse()
        .find((l) => l.trim().startsWith('/**')));
    const docLines = lines.slice(docStartLine, mixinLine);
    const formattedLines = [];
    docLines.forEach((l) => {
        if (l.trim().startsWith('*/')) {
            return;
        }
        if (l.trim().startsWith('/**') && !!l.trim().slice(3).trim()) {
            formattedLines.push(l.trim().slice(3).trim());
        }
        if (l.trim().startsWith('*')) {
            formattedLines.push(l.trim().slice(1).trim());
        }
    });
    const returnStart = formattedLines.findIndex((l) => l.startsWith('@returns'));
    const returnEnd = formattedLines.slice(returnStart + 1).findIndex((l) => l.startsWith('@')) === -1
        ? formattedLines.length - 1
        : formattedLines.slice(returnStart + 1).findIndex((l) => l.startsWith('@')) +
            returnStart;
    const returnLines = formattedLines.slice(returnStart, returnEnd + 1);
    formattedLines.splice(returnStart, returnLines.length);
    const returnType = /@returns *{(\w+)}/.exec(returnLines[0])
        ? /@returns *{(\w+)}/.exec(returnLines[0])[1]
        : '';
    return returnType;
}
exports.extractJsModifierReturnType = extractJsModifierReturnType;
function isInMediaQuery(path) {
    return path.some((n) => n.type === 'atrule' && n.name === 'media');
}
const directives = [
    `-st-from`,
    `-st-named`,
    `-st-default`,
    `-st-root`,
    `-st-states`,
    `-st-extends`,
    `-st-mixin`,
    `-st-partial-mixin`,
    `-st-global`,
];
function isDirective(line) {
    return directives.some((k) => line.trim().startsWith(k));
}
exports.isDirective = isDirective;
function isNamedDirective(line) {
    return line.includes(`-st-named`);
}
function isStImportNamed(line) {
    return line.trim().startsWith(completion_types_1.topLevelDirectives.stImport) && line.includes('[');
}
function isInValue(lineText, position) {
    let isInValue = false;
    if (lineText.includes('value(')) {
        const line = lineText.slice(0, position.character);
        let stack = 0;
        for (let i = 0; i <= line.length; i++) {
            if (line[i] === '(') {
                stack += 1;
            }
            else if (line[i] === ')') {
                stack -= 1;
            }
        }
        if (stack > 0) {
            isInValue = true;
        }
    }
    return isInValue;
}
exports.isInValue = isInValue;
function getChunkAtCursor(fullLineText, cursorPosInLine) {
    let fixedCharIndex = cursorPosInLine;
    let lineChunkAtCursor = fullLineText;
    while (lineChunkAtCursor.lastIndexOf(' ') >= cursorPosInLine) {
        lineChunkAtCursor = lineChunkAtCursor.slice(0, lineChunkAtCursor.lastIndexOf(' '));
    }
    if (!isDirective(lineChunkAtCursor) &&
        lineChunkAtCursor.lastIndexOf(' ') > -1 &&
        lineChunkAtCursor.lastIndexOf(' ') < cursorPosInLine) {
        fixedCharIndex -= lineChunkAtCursor.lastIndexOf(' ') + 1;
        lineChunkAtCursor = lineChunkAtCursor.slice(lineChunkAtCursor.lastIndexOf(' '));
    }
    return { lineChunkAtCursor: lineChunkAtCursor.trim(), fixedCharIndex };
}
function getNamedValues(src, lineIndex) {
    const lines = src.split('\n');
    let isNamedValueLine = false;
    const namedValues = [];
    for (let i = lineIndex; i >= 0; i--) {
        if (isDirective(lines[i]) && !isNamedDirective(lines[i])) {
            break;
        }
        else if (isNamedDirective(lines[i])) {
            isNamedValueLine = true;
            const valueStart = lines[i].indexOf(':') + 1;
            const value = lines[i].slice(valueStart);
            value
                .split(',')
                .map((x) => x.trim())
                .filter((x) => x !== '')
                .forEach((x) => namedValues.push(x));
            break;
        }
        else if (isStImportNamed(lines[i])) {
            isNamedValueLine = true;
            const valueStart = lines[i].indexOf('[') + 1;
            const valueEnd = lines[i].indexOf(']');
            const value = lines[i].slice(valueStart, valueEnd);
            value
                .split(',')
                .map((x) => x.trim())
                .filter((x) => x !== '')
                .forEach((x) => namedValues.push(x));
            break;
        }
        else {
            const valueStart = lines[i].indexOf(':') + 1;
            const value = lines[i].slice(valueStart);
            value
                .split(',')
                .map((x) => x.trim())
                .filter((x) => x !== '')
                .forEach((x) => namedValues.push(x));
        }
    }
    return { isNamedValueLine, namedValues };
}
exports.getNamedValues = getNamedValues;
function getExistingNames(lineText, position) {
    const valueStart = lineText.includes(completion_types_1.topLevelDirectives.stImport)
        ? lineText.indexOf('[') + 1
        : lineText.indexOf(':') + 1;
    const value = lineText.slice(valueStart, position.character);
    const parsed = (0, postcss_value_parser_1.default)(value.trim());
    const names = parsed.nodes
        .filter((n) => n.type === 'function' || n.type === 'word')
        .map((n) => n.value);
    const rev = parsed.nodes.reverse();
    const lastName = parsed.nodes.length && rev[0].type === 'word' ? rev[0].value : '';
    return { names, lastName };
}
exports.getExistingNames = getExistingNames;
function findNode(nodes, index) {
    return nodes
        .filter((n) => n.sourceIndex <= index)
        .reduce((m, n) => {
        return m.sourceIndex > n.sourceIndex ? m : n;
    }, { sourceIndex: -1 });
}
function getDefSymbol(src, position, filePath, stylable) {
    var _a;
    const res = fixAndProcess(src, position, filePath);
    let meta = res.processed.meta;
    if (!meta) {
        return { word: '', meta: null };
    }
    const parsed = (0, postcss_value_parser_1.default)(res.currentLine).nodes;
    let val = findNode(parsed, position.character);
    while (val.nodes && val.nodes.length > 0) {
        if (findNode(val.nodes, position.character).sourceIndex >= 0) {
            val = findNode(val.nodes, position.character);
        }
        else {
            break;
        }
    }
    let word = val.value;
    // sanitize @st-import named imports
    if (word.startsWith('[')) {
        word = word.slice(1, word.length);
    }
    if (word.endsWith(']')) {
        word = word.slice(0, word.length - 1);
    }
    const { lineChunkAtCursor } = getChunkAtCursor(res.currentLine.slice(0, val.sourceIndex + val.value.length), position.character);
    const directiveRegex = new RegExp(`-st-extends` + '|' + `-st-named` + '|' + `-st-default` + '|' + `-st-mixin`);
    if (lineChunkAtCursor.startsWith(':global')) {
        return { word: ':global(' + word + ')', meta };
    }
    const match = lineChunkAtCursor.match(directiveRegex);
    const localSymbol = meta.getSymbol(word);
    if (match && localSymbol) {
        // We're in an -st directive
        let imp;
        if (localSymbol._kind === 'import' && localSymbol.type !== 'default') {
            imp = stylable.resolver.resolveImport(localSymbol);
        }
        else if (localSymbol._kind === 'import' && localSymbol.type === 'default') {
            imp = stylable.resolver.resolveImport(localSymbol);
            return { word: ((_a = imp === null || imp === void 0 ? void 0 : imp.meta) === null || _a === void 0 ? void 0 : _a.root) || '', meta: (imp === null || imp === void 0 ? void 0 : imp.meta) || null };
        }
        else if (localSymbol._kind === 'element' && localSymbol.alias) {
            imp = stylable.resolver.resolveImport(localSymbol.alias);
        }
        else if (localSymbol._kind === 'class') {
            if (localSymbol.alias) {
                const res = stylable.resolver.resolveImport(localSymbol.alias);
                return { word, meta: res ? res.meta : null };
            }
            return { word, meta };
        }
        if (imp) {
            if (imp._kind === 'js') {
                return { word, meta };
            }
            else {
                return { word, meta: imp.meta };
            }
        }
        else {
            return { word: '', meta: null };
        }
    }
    const varRegex = new RegExp('value\\(\\s*' + word);
    if (varRegex.test(lineChunkAtCursor)) {
        // we're looking at a var usage
        const symbol = meta.getSymbol(word);
        if (!symbol) {
            return { word, meta: null };
        }
        else if (symbol._kind === 'var') {
            // deepResolve doesn't do local symbols
            return { word, meta };
        }
        const resolvedVar = stylable.resolver.deepResolve(symbol);
        if (resolvedVar) {
            return { word, meta: resolvedVar.meta };
        }
        else {
            return { word, meta: null };
        }
    }
    const expandedLine = (0, core_1.expandCustomSelectors)(postcss.rule({ selector: lineChunkAtCursor }), meta.customSelectors)
        .split(' ')
        .pop(); // TODO: replace with selector parser
    const resolvedElements = stylable.transformSelector(meta, expandedLine).resolved;
    let reso;
    if (!word.startsWith(word.charAt(0).toLowerCase())) {
        reso = resolvedElements[0][resolvedElements[0].length - 1].resolved.find((res) => !!res.symbol['-st-root']);
    }
    else if (resolvedElements.length && resolvedElements[0].length) {
        reso = resolvedElements[0][resolvedElements[0].length - 1].resolved.find((res) => {
            let symbolStates;
            if (res.symbol._kind === 'class') {
                symbolStates = res.symbol[`-st-states`];
            }
            return ((res.symbol.name === word.replace('.', '') && !res.symbol.alias) ||
                (symbolStates && Object.keys(symbolStates).some((k) => k === word)));
        });
    }
    if (reso) {
        meta = reso.meta;
    }
    return { word, meta };
}
exports.getDefSymbol = getDefSymbol;
//# sourceMappingURL=provider.js.map